@@ -4,6 +4,11 @@ use parent 'MariaDB::NonBlocking';
 use v5.18.2; # needed for __SUB__, implies strict
 use warnings;
 
+use constant DEBUG => $ENV{MariaDB_NonBlocking_DEBUG} // 0;
+sub TELL (@) {
+    say STDERR __PACKAGE__, ': ', join " ", @_;
+}
+
 BEGIN {
     my $loaded_ok;
     local $@;
@@ -16,7 +21,7 @@ use Promises (); # for deferred
 
 # Better to import this, since it is a custom op
 use Ref::Util qw(is_ref is_arrayref);
-use Scalar::Util qw(weaken);
+use Scalar::Util qw(refaddr weaken);
 
 use AnyEvent;
 
@@ -31,11 +36,10 @@ my $IS_EV;
 # EV lets us cut down the number of watchers created
 # per connection significantly.
 AnyEvent::post_detect {
-    $IS_EV = ($AnyEvent::MODEL//'') eq 'AnyEvent::Impl::EV'
+    $IS_EV = ($AnyEvent::MODEL//'') eq 'AnyEvent::Impl::EV' ? 1 : 0;
+    DEBUG && TELL $IS_EV ? "Using EV" : "Using AnyEvent";
 };
 
-use constant ();
-
 BEGIN {
     my $loaded_ok;
     local $@;
@@ -47,11 +51,11 @@ BEGIN {
 
     if ( $loaded_ok ) {
         # EV loaded fine.  Add some aliases
-        *EV_READ  = \*EV::READ;
-        *EV_WRITE = \*EV::WRITE;
-        *EV_TIMER = \*EV::TIMER;
-        *EV_io    = \*EV::io;
-        *EV_timer = \*EV::timer;
+        *EV_READ       = \*EV::READ;
+        *EV_WRITE      = \*EV::WRITE;
+        *EV_TIMER      = \*EV::TIMER;
+        *EV_io         = \*EV::io;
+        *EV_timer      = \*EV::timer;
         *EV_now_update = \*EV::now_update;
     }
     else {
@@ -154,6 +158,8 @@ sub __grab_watcher {
         Carp::confess("Overriding a $watcher_type watcher!  How did this happen?");
     }
 
+    AnyEvent::detect() unless defined $IS_EV;
+
     if ( $IS_EV ) {
         # If we are using EV, reuse a watcher if we can.
         my $cb = __wrap_ev_cb($watcher_args->[2]);
@@ -168,6 +174,7 @@ sub __grab_watcher {
             if ( !$existing_watcher ) {
                 # No pre-existing watcher for us to use;
                 # make a new one!
+                DEBUG && TELL "Started new $watcher_type watcher ($watcher_args->[1])";
                 $storage->{$watcher_type} = EV_io(
                     $watcher_args->[0],
                     $ev_mask,
@@ -176,6 +183,7 @@ sub __grab_watcher {
                 return;
             }
 
+            DEBUG && TELL "Reusing existing $watcher_type watcher ($watcher_args->[1])";
             $existing_watcher->set(
                     $watcher_args->[0],
                     $ev_mask,
@@ -184,6 +192,7 @@ sub __grab_watcher {
         elsif ( index($watcher_type, 'timer') != -1 ) {
             EV_now_update();
             if ( !$existing_watcher ) {
+                DEBUG && TELL "Started new $watcher_type watcher";
                 $storage->{$watcher_type} = EV_timer(
                     $watcher_args->[0],
                     $watcher_args->[1],
@@ -191,6 +200,7 @@ sub __grab_watcher {
                 );
                 return;
             }
+            DEBUG && TELL "Reusing existing $watcher_type watcher";
             $existing_watcher->set(
                     $watcher_args->[0],
                     $watcher_args->[1],
@@ -210,8 +220,11 @@ sub __grab_watcher {
         # Slightly easier, really.  Less performant since
         # we never reuse watchers.
         if ( index($watcher_type, 'timer') != -1 ) {
+            delete $storage->{$watcher_type};
+
             my $cb         = $watcher_args->[2];
             my $wrapped_cb = sub { return $cb->(MYSQL_WAIT_TIMEOUT) };
+            DEBUG && TELL "Started new $watcher_type watcher";
             AnyEvent->now_update;
             $storage->{$watcher_type} = AnyEvent->timer(
                 after    => $watcher_args->[0],
@@ -231,6 +244,7 @@ sub __grab_watcher {
             # AnyEvent works around it though, so all is good.
             my $wait_for      = $watcher_args->[1];
             my $cb            = $watcher_args->[2];
+            DEBUG && TELL "Started new $watcher_type watcher ($wait_for)";
             $storage->{io_r} = AnyEvent->io(
                 fh   => $watcher_args->[0],
                 poll => "r",
@@ -329,9 +343,10 @@ sub ____run {
 
     my $all_watchers = {};
     CONNECTION:
-    foreach my $outside_maria ( @$connections ) {
+    foreach my $connection_idx ( 0..$#{$connections} ) {
         last CONNECTION if @errors;
 
+        my $outside_maria = $connections->[$connection_idx];
         my $wait_for = $call_start->($outside_maria);
 
         if ( !$wait_for ) {
@@ -358,6 +373,8 @@ sub ____run {
         my $maria = $outside_maria;
         weaken($maria);
 
+        my $deferred_refaddr = refaddr($deferred);
+
         my $watcher_ready_cb = sub {
             if ( $extras->{cancel} ) {
                 # Right... this needs some explanation.
@@ -393,6 +410,7 @@ sub ____run {
                 # then we just return from this loop; otherwise we
                 # reject it.
                 __return_all_watchers_for_multiple_sockets($all_watchers);
+                delete $maria->{pending}{$deferred_refaddr};
                 $deferred->reject("Manual cancellation, reason: $extras->{cancel}")
                     if $deferred->is_in_progress;
                 return;
@@ -404,6 +422,7 @@ sub ____run {
             }
 
             if ( !$maria ) {
+                delete $maria->{pending}{$deferred_refaddr};
                 $deferred->reject("Connection object went away")
                     if $deferred->is_in_progress;
                 __return_all_watchers_for_multiple_sockets($all_watchers);
@@ -449,6 +468,7 @@ sub ____run {
                     # Destroy ALL watchers, otherwise
                     # we might leak some!
                 # We got an error above.  Reject the promise and bail
+                delete $maria->{pending}{$deferred_refaddr};
                 $deferred->reject(@errors);
                 return;
             }
@@ -460,6 +480,7 @@ sub ____run {
                 __return_all_watchers(delete $all_watchers->{$socket_fd}); # BOI!!
                 if ( !keys %$all_watchers ) {
                     # Ran all the queries! We can resolve and go home
+                    delete $maria->{pending}{$deferred_refaddr};
                     $deferred->resolve(\@per_query_results);
                     return;
                 }
@@ -533,16 +554,21 @@ sub ____run {
                 $perl_timeout,
                 0, # no repeat
                 sub {
+                    DEBUG && TELL "Global timeout reached";
                     __return_all_watchers_for_multiple_sockets(
                         $all_watchers
                     );
 
                     push @errors,
                         "$type execution was interrupted by perl, maximum execution time exceeded (timeout=$perl_timeout)";
+                    delete $maria->{pending}{$deferred_refaddr};
                     $deferred->reject(@errors);
                 },
             ]
         }) if $perl_timeout;
+
+        $maria->{pending}{$deferred_refaddr} = $deferred;
+        weaken($maria->{pending}{$deferred_refaddr});
     }
 
     my $promise = $deferred->promise;
@@ -562,6 +588,19 @@ sub ____run {
     return $promise;
 }
 
+sub DESTROY {
+    my $self = shift;
+
+    my $pending = $self->{pending} // {};
+    my $pending_num = 0;
+    for my $deferred ( grep defined, values %$pending ) {
+        next unless $deferred->is_in_progress;
+        $pending_num++;
+        $deferred->reject("Connection object went away");
+    }
+    DEBUG && $pending_num && TELL "Had $pending_num operations still running when we were freed";
+}
+
 sub run_multiple_queries {
     my ($conns, $remaining_sqls, $extras) = @_;
 
